// 
// GameX -- CameraX Demo
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#define GAMEX_MAIN
#pragma comment (lib, "gamex.lib")
#include "gamex.hpp"
										//*** Global Variables & Defines
#define MAX_PARTICLES 50				// Maximum number of particles we think we'll need
#define PARTICLE_ADD_FREQ 2				// Every PARTICLE_ADD_FREQ cycles, we add a particle
#define MAX_PARTICLE_AGE 100			// After MAX_PARTICLE_AGE, particle is removed
#define PARTICLE_LAT_SPEED_MAX 1.0f		// Maximum lateral speed a particle is added at
#define PARTICLE_VERT_SPEED_MIN 0.25f	// Minimum vertical speed a particle is added at
#define PARTICLE_VERT_SPEED_MAX 2.0f	// Maximum vertical speed a particle is added at
#define GRAVITY_CONST 0.05f				// Added to each particle's downward velocity each cycle
#define TRAIL_LENGTH 32					// Length of trail behind particles
#define TRAIL_RES 13					// Number of trail positions per line segment
#define TERRAIN_ANG_RES 30				// Number of degrees across each terrain segment
#define TERRAIN_DIST_RES 20				// Distance along each terrain segment
#define TERRAIN_DIST_MAX 100			// Maximum distance of terrain segments
#define MAX_POLYS (TERRAIN_DIST_MAX*360/TERRAIN_DIST_RES/TERRAIN_ANG_RES) // Maximum polygons we think we'll need
#define BLUR_FACTOR 1.75f				// 1.0 means full motion blur, 2.0 means wide radial blur, must be >= 1
#define	BLUR_MAX_OPACITY 128			// Max amount of blur to write each frame, should be >= 64 and <= 192

int num_particles = 0, num_polys = 0;	// Total number of particles and polygons currently used
CameraX cam;							// Camera for 3D projections
ImageX particle_img, floor_img, blur_img;	// Sprite of a particle, texture for the floor, and buffer for the blur
bool effectOn [5];						// True for each effect that's turned on; turn off using keys 1-5
int blurOpacity = 0;					// Current amount of blur to write each frame, goes from 0 to BLUR_MAX_OPACITY
float x_angle = 0.0f;					// x angle of our camera
float x_dang = 100.0f;					// change in x angle of camera
float z_dist = 100.0f;					// z distance (zoom) of camera
float z_plus = 2000.0f;					// offset to z distance of camera
float y_angle = 15.0f;					// y angle (pitch) of camera

struct Particle {
	Vector3DF pos[TRAIL_LENGTH];
	Vector3DF vel;
	int age;
	int curr_pos;
};

Particle particle [MAX_PARTICLES];

struct StoredPoly {
	Vector3DF v[4];
	ColorX c;
};

StoredPoly poly [MAX_POLYS];

// define functions:
void calculatePolygons (void);

void addParticles (void);
void updateParticles (void);
void setCameraAngles (void);
void updateBlurAmount (void);
void enableOrDisableEffects (void);

void drawParticles (void);
void drawTerrain (void);
void drawBackground (void);
void doRadialBlur (void);

float getHeight (Vector3DF & v);
Vector3DF radialToPos (float dist, float ang);


void GameRun (void)
{	
	addParticles();
	updateParticles();
	enableOrDisableEffects();
	setCameraAngles();
	updateBlurAmount();
}

void GameDraw (void)
{
	drawBackground();

	// Normally we would call GameX.Begin3DScene() here, but that tells GameX to sort the scene,
	// and we don't want GameX to sort this particular scene because we draw it in a custom order.
	// So, we'll just start drawing the scene without calling anything else first:

	drawTerrain();
	drawParticles();

	// Now that we're done with our scene, we would normally call GameX.End3DScene() here,
	// but since we didn't call Begin3DScene, we shouldn't call End3DScene either.

	doRadialBlur(); // if user hit the 5 key, draw the radial blur
					// note that anything drawn after the blur does not get blurred:

	if(GameX.GetGlobalCounter() < 500) // if the game has gone through less than 500 run loops,
		GameX.DrawOutlinedText(10,10, "Use the keys 1 through 5 to enable/disable effect layers. Use the mouse to change the camera angle.");
}

void GameInit (void)
{											// Initialize GameX
	GameX.Initialize ("ParticleDemoX", VIDEO_16BIT|VIDEO_ALLOWREFRESHSYNC|RUN_ALWAYSHIDECURSOR, 1024, 768);
	
	cam.SetWindow (0, 0, GameX.GetWidth(), GameX.GetHeight());	// Initialize Camera
	cam.SetToPosition (Vector3DF(0, 0, 0)); // Set camera to look at the origin
	GameX.SetDrawCam (&cam); // Tell GameX to draw using the camera
	
	particle_img.Load ("Images\\photon.bmp"); // note that neither of our images have an alpha channel;
	floor_img.Load ("Images\\alien.bmp");  // alpha in the image is not needed for all types of transparency
	blur_img.Create ((int)((float)GameX.GetWidth()/BLUR_FACTOR), (int)((float)GameX.GetHeight()/BLUR_FACTOR)); // (create image for radial blur)

	calculatePolygons();

	for(int i = 0 ; i < 5 ; i++)
		effectOn[i] = true; // turn effects on by default
	effectOn[4] = false; // except start with radial blur off
}

void drawBackground (void)
{
	// draw a green-to-black rectangle above the horizon line, and an all-black rectangle below that:
	float horizonRatio = 50.0f/z_dist + 0.25f - y_angle/180.0f;
	GameX.DrawRect(ColorX(0,128,64,255, 0,128,64,255, 0,0,0,255, 0,0,0,255), 0,0,GameX.GetWidth(),GameX.GetHeight()*horizonRatio);
	GameX.DrawRect(ColorX(0,0,0), 0,GameX.GetHeight()*horizonRatio,GameX.GetWidth(),GameX.GetHeight());
}

void drawTerrain (void)
{
	if(!effectOn[0]) return;

	// just loop through and draw the polygons
	for(int p = 0 ; p < num_polys ; p++) {
		GameX.SetDrawMode(DRAWOP_NOCULL); // make these polygons 2-sided -- DON'T DO THIS NORMALLY, it's just for demo purposes here
		GameX.SetDrawShading(poly[p].c); // precalculated shading
		GameX.DrawTexturedPolygon3D(&floor_img, poly[p].v[0],poly[p].v[1],poly[p].v[2],poly[p].v[3]);
	}
}

void drawParticles (void)
{
	// Draw 3 layers of effects for each particle,
	// note that since all of these use DRAW_ADD, order doesn't matter here (A+B == B+A)

	for (int n = 0; n < num_particles; n++) {
		float intensity = ((float)(MAX_PARTICLE_AGE - particle[n].age)/(float)MAX_PARTICLE_AGE);

		// *** Draw the point-sprite particle effect:
		// (the simplest effect, just draw a sprite at the location of each particle)

		if(effectOn[1]) {
			GameX.SetDrawScale (16.0f); // make each particle 16 units across
			GameX.SetDrawAngle ((float)(-15*GameX.GetGlobalCounter())); // spinning quickly
			GameX.SetDrawMode (DRAW_ADD|DRAWOP_CENTERED); // additive special effect and centered at point
			GameX.SetDrawShading (ColorX(64,192,255) * intensity); // blue-ish color... it's important to make each component at least 32 (when at full intensity)
			GameX.DrawImage3D (&particle_img, particle[n].pos[particle[n].curr_pos]); // project and draw the particle
		}

		// *** Draw the line-sprite particle trail effect:
		// (the slowest part, since we draw each particle's trail using many line segments to allow curving)

		if(effectOn[2]) {
			for(int j = 0 ; j < TRAIL_LENGTH-(2+TRAIL_RES) ; j+=4) {
				int i = (j+1+particle[n].curr_pos)%TRAIL_LENGTH;
				int i2 = (j+1+TRAIL_RES+particle[n].curr_pos)%TRAIL_LENGTH;
				GameX.SetDrawMode (DRAW_ADD);
				GameX.SetDrawShading (ColorX(96,128,255) * (intensity * ((float)j/(float)(TRAIL_LENGTH-(2+TRAIL_RES)))));
				GameX.DrawLineImage3D (&particle_img, particle[n].pos[i], particle[n].pos[i2], true); // project and draw the line
			}
		}

		// *** Draw the polygonal particle lighting effect:
		// (normally we would have to elevate these a little to make sure they draw over the ground, but they're not being sorted)

		if(effectOn[3]) {
			int curr_pos = particle[n].curr_pos;
			float min = getHeight(particle[n].pos[curr_pos]);
			float bright = min(2.0f, 2.0f*(10.0f - particle[n].pos[curr_pos].z+min)/10.0f);
 			if(bright >= 0.0f) {
				Vector3DF groundPoint = Vector3DF(particle[n].pos[curr_pos].x,particle[n].pos[curr_pos].y,0);
				Vector3DF dx = Vector3DF(10,0,0)*bright;
				Vector3DF dy = Vector3DF(0,10,0)*bright;
				Vector3DF v [4] = {groundPoint-dx-dy, groundPoint-dx+dy, groundPoint+dx-dy, groundPoint+dx+dy};
				for(int i = 0 ; i < 4 ; i++) v[i].z = getHeight(v[i]);
				GameX.SetDrawMode (DRAW_ADD|DRAWOP_NOCULL); // makes lighting polygons two-sided
				GameX.SetDrawShading (ColorX(128,64,255) * (bright * intensity));
				GameX.DrawTexturedPolygon3D (&particle_img, v[0],v[1],v[2],v[3]); // project and draw the polygon
			}
		}
	}
}
void doRadialBlur (void)
{
	if(blurOpacity > 0) { // radial blur:
		// this is where we would do exotic things like set rotation, scaling, DRAW_ADD, etc.
		// but we'll just do a perfectly normal radial blur here
		GameX.SetDrawShading (ColorX(255,255,255,blurOpacity)); // blurOpacity determines intensity of blur
		GameX.DrawTexturedRect (&blur_img, 0,0,GameX.GetWidth(),GameX.GetHeight()); // layer the blur over the screen

		// the following takes a central chunk of the screen that is the same size as blur_img,
		// and copies that part of the screen directly into blur_img for use above next time:
		// (it's important for this to happen AFTER the above code in order to get a feedback effect)
		int xBorder = (GameX.GetWidth() - blur_img.GetWidth())/2;
		int yBorder = (GameX.GetHeight() - blur_img.GetHeight())/2;
		GameX.SetDrawPart (xBorder,yBorder,GameX.GetWidth()-xBorder,GameX.GetHeight()-yBorder);
		GameX.SetDrawDestination (&blur_img);
		GameX.DrawImage (VIEWPORT, 0,0); // this must be drawn with no scaling or rotation to be fast
	}
}

void addParticles (void)
{
	// Add another particle if it's time to do so and we're under the limit
	if(GameX.GetGlobalCounter() % PARTICLE_ADD_FREQ == 0 && num_particles < MAX_PARTICLES) {
		for(int i = 0 ; i < TRAIL_LENGTH ; i++)
			particle[num_particles].pos[i].Set(0,0,0);
		particle[num_particles].vel.Set(GameX.GetRandomFloat(-PARTICLE_LAT_SPEED_MAX,PARTICLE_LAT_SPEED_MAX),
								   GameX.GetRandomFloat(-PARTICLE_LAT_SPEED_MAX,PARTICLE_LAT_SPEED_MAX),
								   GameX.GetRandomFloat(PARTICLE_VERT_SPEED_MIN,PARTICLE_VERT_SPEED_MAX));
		particle[num_particles].age = 0;
		particle[num_particles].curr_pos = 0;
		num_particles++;
	}
}

void updateParticles (void)
{
	// Move particles, age them, and remove when too old
	for (int n = 0; n < num_particles; n++) {
		particle[n].vel.z -= GRAVITY_CONST;
		int curr_pos = particle[n].curr_pos;
		int new_pos = (particle[n].curr_pos+1)%TRAIL_LENGTH;
		particle[n].pos[new_pos] = particle[n].pos[curr_pos] + particle[n].vel;
		particle[n].curr_pos = new_pos;
		float min = getHeight(particle[n].pos[new_pos]);
		if(particle[n].pos[new_pos].z <= min) {
			particle[n].pos[new_pos].z = min+min-particle[n].pos[new_pos].z;
			particle[n].vel.z = -0.6f * particle[n].vel.z;
		}
		particle[n].age++;
		if(particle[n].age > MAX_PARTICLE_AGE) {
			memcpy (&particle[n], &particle[--num_particles], sizeof(Particle)); // last one in its place.
		}
	}
}

void setCameraAngles (void)
{
	// Calculate the camera's angles and distance from where it's looking at:
	x_dang *= 0.95f;
	x_dang += (float)GameX.GetMouseDX() / 50.0f;
	z_plus *= 0.99f;
	z_plus += (float)GameX.GetMouseDY() / 5.0f - GameX.GetMouseDZ();
	if(z_plus < -75.0f) z_plus = -75.0f;
	if(z_plus > 2000.0f) z_plus = 2000.0f;
	x_angle += x_dang + 0.5f;
	y_angle = GameX.GetOscillatingValue (15.0f, 45.0f, 1.0f);
	z_dist  = GameX.GetOscillatingValue (100.0f, 200.0f, 0.5f) + z_plus + (float)(blurOpacity*blurOpacity/100);

	cam.SetFromAngles (Vector3DF(x_angle, y_angle, z_dist));
}

void updateBlurAmount (void)
{
	// adjust the blur amount a little at a time so the blur doesn't suddenly snap into effect:
	if(effectOn[4]) {
		if(blurOpacity < BLUR_MAX_OPACITY)
			blurOpacity++;
	} else {
		if(blurOpacity > 0)
			blurOpacity--;
	}
}

void enableOrDisableEffects (void)
{
	// Turn on/off effects when the user presses keys 1-5:
	for(int i = 0 ; i < 5 ; i++)
		if(GameX.GetKeyPress(KEY_1 + i))
			effectOn[i] = !effectOn[i];
}

// returns the height on the ground at the given x,y coords
float getHeight (Vector3DF & v)
{
	return -0.005f * (v.x*v.x + v.y*v.y) + 0.000001f * (v.x*v.x*v.x*v.x + v.y*v.y*v.y*v.y);
}

// returns the position (x,y,z) vector that the given radial (r,theta) coords translate to
Vector3DF radialToPos (float dist, float ang)
{
	float x = dist * cosf(ang*DEGtoRAD);
	float y = dist * sinf(ang*DEGtoRAD);
	Vector3DF v = Vector3DF(x,y,0);
	v.z = getHeight(v);
	return v;
}

void calculatePolygons (void)
{
	// precalculate polygon locations and colors since we don't need them to change during the scene:
	float ang, dist;
	for(dist = 0 ; dist < TERRAIN_DIST_MAX ; dist+=TERRAIN_DIST_RES) {
		for(ang = 0 ; ang < 360 ; ang += TERRAIN_ANG_RES) {
			if(num_polys >= MAX_POLYS) return;
			int sector = (int)(ang/TERRAIN_ANG_RES);
			int bright = (int)((TERRAIN_DIST_MAX-dist)/(float)TERRAIN_DIST_MAX*255.0f);
			int bright2 = (int)((TERRAIN_DIST_MAX-(dist+TERRAIN_DIST_RES))/(float)TERRAIN_DIST_MAX*255.0f);
			poly[num_polys].v[0] = radialToPos (dist, ang);
			poly[num_polys].v[1] = radialToPos (dist, ang+TERRAIN_ANG_RES);
			poly[num_polys].v[2] = radialToPos (dist+TERRAIN_DIST_RES, ang);
			poly[num_polys].v[3] = radialToPos (dist+TERRAIN_DIST_RES, ang+TERRAIN_ANG_RES);

			poly[num_polys].c = ColorX(bright,bright/2,255,bright,   bright,bright/2,255,bright,   bright2,bright2/2,255,bright2,   bright2,bright2/2,255,bright2);
			// Note we are giving our polygons transparency,
			// which is normally a BAD IDEA, but since we are sorting these polygons ourselves
			// and since it's a demo with not very many polygons anyway, it's ok here

			num_polys++;
		}
	}
}